//  Using the analytic center in the Feasibility Pump
//
//  usage:
//  feaspump maxit1 maxit2 maxit1wi maxit2wi minflip trsmin trsrnd algo s3me lppresol mippresol imp alphainit alphaquot alphadist epsfeas max_time solver y lpmodel\n"
//          "    maxit1    = 0 to disable stage1 or max iterations for stage 1 (10000)\n"
//          "    maxit2    = 0 to disable stage2 or max iterations for stage 2 (2000)\n"
//          "    maxit1wi  = max stage1 iterations without 10% improvement (70)\n"
//          "    maxit2wi  = max stage2 iterations without 10% improvement (600)\n"
//          "    minflp    = minimum # of variables to flip per iteration (20)\n"
//          "    trsmin    = integrality threshold below which variables are not flipped (.001)\n"
//          "    trsrnd    = s|r|h|q|a : rounding threshold (standard 1/2, random in 0-1,\n" 
//          "                random in .25-.75, random with quadratic distribution, auto) (a)\n"
//          "    algo      = p|d|b|a (primal, dual, barrier, auto)\n"
//          "    s3me      = stage 3 cplex mip emphasis (4)\n"
//          "    lppresol  = y|n: use presolving on LPs (y)\n"
//          "    mippresol = y|n: use presolving on MIP prior to invoking the feaspump (y)\n"
//          "    imp       = % of the gap to cut off (0 to disable improvement phase; negative\n"
//          "                to disable even improvement with fixed generals) (25)\n"
//          "    alphainit = start value of the original objective function fraction (1.0)\n"
//          "    alphaquot = redunction factor of alpha (0.9)\n"
//          "    alphadist = difference up to which two alpha values are considered equal (0.005)\n"
//	    "    epsfeas   = feasibility threshold in order to decide if a new point is (or not) feasible (lb <= Ax <= ub) \n"
//	    "    max_time = max time for heuristic FP\n"
//	    "    solver = solver used to find analytic center (1=PCx, 2=GLPK, 3=CPLEX)\n"
//	    "    norminf = n = standard FP algorithm if no feasible solution between RP-AC, y= infinity norm (integer point less infeasible)\n"
//	    "\n"
//          "    e.g.: feaspump 10000 2000 70 600 20 .02 q a 4 n y -1 1.0 0.9 0.005 1.0e-6 7200 1 y <lpmodel>\n"
//
//
//  Authors:
//  Daniel Baena
//  Dept. of Statistics and Operations Research
//  Universitat Politecnica de Catalunya
//  daniel.baena@upc.edu
//
//  Jordi Castro
//  Dept. of Statistics and Operations Research
//  Universitat Politecnica de Catalunya
//  jordi.castro@upc.edu
//
//
//
//
//
// Feasibility Pump
// Proof-of-concept implementation with ILOG Cplex 9/Concert Technology
//
// Reference papers:
//
// A Feasibility Pump heuristic for general Mixed-Integer Problems
// L. Bertacco, M. Fischetti, A. Lodi. 
// Technical Report OR/05/5 DEIS Universita' di Bologna, 2005
// 
// The Feasibility Pump. 
// M. Fischetti, F. Glover, A. Lodi. 
// Mathematical Programming 2005
// Digital Object Identi�er (DOI) 10.1007/s10107-004-0570-3.
//
// Improving the Feasibility Pump
// T. Achterberg, T. Berthold
// Technical Report ZR-05-42, Konrad-Zuse-Zentrum fuer Informationstechnik Berlin, 2005
//
// Livio Bertacco (livio@bertacco.it), July 8th, 2005
// PhD Candidate, University of Padua
// All rights reserved.


#include <string>
#include <fstream>
#include <iostream>
#include <sstream>
#include <limits>
#include <iomanip>
#include <vector>
#include <set>
#include <ext/hash_set>
#include <ext/hash_map>
#include <time.h>

using namespace std;

#include <ilcplex/ilocplex.h>
#include <ilconcert/ilorandom.h>
#include "fp.h"
#include "AC.h"
using namespace __gnu_cxx;

IloEnv env;		               // Cplex environment
IloRandom rng(env);                    // random number generator
IloCplex cplex(env);		       // Cplex solver

IloModel model(env);		       // Cplex model
IloExpr origObjExpr(env);	       // expression of origObj
IloObjective activeObj(env);	       // current obj func.




IloExpr ACExpr(env);		       // ACExpr=0 to find the analytic center with barrier algorithm without crossover(IP method)
hash_map <const char*,int> index_of_vars;
vector<bool> isInt;		       	// for each variable of the problem I know if is integer (or binary) or continue
IloNumArray relaxedVars(env);		// value of the vars (all) in the current relaxation
double *ACroundedIntVars;
double lbinfeas = numeric_limits<double>::infinity(); // lower bound infeasibility
double currentinfeas = -numeric_limits<double>::infinity(); // current infeasibility (infinity norma vector)
double bestgamma=1;			// best gamma. gamma associated to best feasible point
double lessgamma=0;			// gamma associated to point less feasible

double foconst;			       // constant term in the obj function
IloNumVarArray vars(env);	       // set of all problem vars
IloRangeArray rngs(env);	       // set of all problem constraints
IloNumVarArray intVars(env);	       // set of binary variables (subset of vars)
IloNumArray relaxedIntVars(env);       // value of the int vars in the current relaxation
vector<double> roundedIntVars;	       // value of the int vars in the current integer aproximation 
vector<double> prevRoundedIntVars;
int s1restarts = 0;
IloExpr dist(env);
double origobj_norm;                   // norm of the objective function 
double alpha;                          // ratio by which the new obj is combined from distance and original obj
double alphadist;                      // difference up to which two alpha values are considered equal

template<class _T>
class hash_compare_for_pairs {
private:
   typedef pair<vector<_T>,_T> Key;
public:
   static const size_t bucket_size = 4;
   static const size_t min_buckets = 20000;
   size_t operator()(const Key & k) const {
      size_t h = 0;
      for( int i = 0; i<k.first.size(); i++ ) 
         h = ((h<<1) | ((h>>31) & 1)) ^ (long)k.first[i];
      return h;
   };
};

template<class _T>
class equalkey_for_pairs {
private:
   typedef pair<vector<_T>,_T> Key;
public:
   bool operator()(const Key & k1, const Key & k2) const
   {
      return (k1.first == k2.first) && (fabs(k1.second-k2.second) <= alphadist);
   };
};

hash_set<pair<vector<double>,double>, hash_compare_for_pairs<double>, equalkey_for_pairs<double> > history; 


vector<bool> isBin;

IloNumVar** deltaVars;
IloRange** deltaCnstrUB;
IloRange** deltaCnstrLB;

double epInt;					// integer tolerance
int minChange;					// minimum number of flipped var between consecutive integer points
int maxIter1, maxIter2;				// maximum number of iterations
int nVars;					// number of variables in the problem
int nBinVars;					// number of binary variables in the problem
int nIntVars;					// number of integer (and binary) variables in the problem
int changed;					// number flipped variables
int nDeltaVars = 0;
int restarts = 0;
int stage;
int iter = 0;
double trsmin;
int trsrnd;
double obj_offset = 0;

IloCplex auxcplex(env);
IloModel auxmodel(env);
IloObjective auxobj(env);		
IloNumVarArray auxvars(env);	
IloRangeArray auxrngs(env);		
IloNumVarArray auxintvars(env);

static double getFractionality()
{
   double fracsum = 0.0;
   for( int i = 0; i < nIntVars; i++ )
   {
      double roundsol = floor(relaxedIntVars[i] + 0.5);
      double frac = fabs(roundsol - relaxedIntVars[i]);
      fracsum += frac;
   }
   return fracsum;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
//Auxiliar functions
////////////////////////////////////////////////////////////////////////////////////////////////////

// relation between name of each variable and their Id
void info_pos_of_each_variable(int nVars) 
{
	for (int i=0;i<nVars;i++) index_of_vars[vars[i].getName()]=i;
}


double getFObjValue (double* IP)
{
	IloExpr::LinearIterator it = origObjExpr.getLinearIterator();
	double objvalue=0.0;
	while (it.ok())
	{
		objvalue += it.getCoef()*IP[index_of_vars[it.getVar().getName()]];
		++it;
	}
return objvalue;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
//feasibility 
//given a point x tell us if feasible or infeasible (Ax=b)
/////////////////////////////////////////////////////////////////////////////////////////////////////

bool isFeasible(double *x,int nVars,int nRows,double epsfeas)
{
	IloExpr e;
	double coef;
	bool feas = true;
	double AX=0.0;
	int i=0;
	double lb,ub;
	double infeasible;
	
	while ((i<nRows))
	{
		e=rngs[i].getExpr();
		ub=rngs[i].getUB();
		lb=rngs[i].getLB();
	
		IloExpr::LinearIterator	it = e.getLinearIterator();
		while (it.ok())
		{
			AX=AX + (it.getCoef()* x[index_of_vars[it.getVar().getName()]]);
			++it;
		}//end while
	
		//check the feasibility and save the value of most infeasible vector 
		if (((AX + (1+ fabs(AX))*epsfeas) < lb) || ((AX- (1+ fabs(AX))*epsfeas) > ub)) 
		{
			feas = false;
			infeasible= min(fabs((AX + (1+ fabs(AX))*epsfeas-lb)),fabs((AX - (1+ fabs(AX))*epsfeas-ub)));
			if (infeasible > currentinfeas) currentinfeas=infeasible;
		}
		i++;
		AX=0.0;
	
	} // end while
	
	if (feas) {cout << "... is FEASIBLE point " << endl;}
	//else {cout << "... is INFEASIBLE point " << endl; }
	
return feas;
}



int main(int argc, char* argv[])
{
   double *analcent;  // value of the analytic center.	
   double *InitialRP;
   double *InteriorPoint;
   double dbl_infinity = numeric_limits<double>::infinity();

   bool found = false;
   
   double lastObj = 0;
   char *lpfile = 0;
   char *modelfile = 0;
   int nRows;	

   int ext_iter = 0;
   int tot_iter = 0;
   int tot_restarts = 0; 
   int tot_found = 0;
   double lp_bound;
   double best_sol;
	
   int	first_iter;
   double first_sol = 0;
   int first_restarts;
   int first_stage = 0;
   double imp;
   int maxnamelen = 0;

   time_t timestart;
   time_t timeend;
   time_t cplex_time_start;
   time_t cplex_time_end;
   int cplex_time = 0;

   
   time_t stageAC_time_start;
   time_t stageAC_time_end;
   time_t stage0_time_start;
   time_t stage0_time_end;
   time_t stage1_time_start;
   time_t stage1_time_end;
   time_t stage2_time_start;
   time_t stage2_time_end;
   time_t stage3_time_start;
   time_t stage3_time_end;
   
   
   bool removetmppresolfile = false;
   char tmppresolfile[] = "fp_tmp_presolved.pre";

   try {
      timestart = time(NULL);
      cout << "FPgen-t 06/06/2005 - Bertacco, Fischetti, Lodi" << endl;
      cout << "        09/16/2005 - Achterberg, Berthold" << endl;
      cout << "        01/03/2010 - Baena, Castro\n" << endl;
        
      if( argc != 21 ) 
         throw string(
            "Usage: feaspump maxit1 maxit2 maxit1wi maxit2wi minflip trsmin trsrnd algo s3me lppresol mippresol imp alphainit alphaquot alphadist epsfeas max_time solver norminf lpmodel\n"
            "    maxit1    = 0 to disable stage1 or max iterations for stage 1 (10000)\n"
            "    maxit2    = 0 to disable stage2 or max iterations for stage 2 (2000)\n"
            "    maxit1wi  = max stage1 iterations without 10% improvement (70)\n"
            "    maxit2wi  = max stage2 iterations without 10% improvement (600)\n"
            "    minflp    = minimum # of variables to flip per iteration (20)\n"
            "    trsmin    = integrality threshold below which variables are not flipped (.001)\n"
            "    trsrnd    = s|r|h|q|a : rounding threshold (standard 1/2, random in 0-1,\n" 
            "                random in .25-.75, random with quadratic distribution, auto) (a)\n"
            "    algo      = p|d|b|a (primal, dual, barrier, auto)\n"
            "    s3me      = stage 3 cplex mip emphasis (4)\n"
            "    lppresol  = y|n: use presolving on LPs (y)\n"
            "    mippresol = y|n: use presolving on MIP prior to invoking the feaspump (y)\n"
            "    imp       = % of the gap to cut off (0 to disable improvement phase; negative\n"
            "                to disable even improvement with fixed generals) (25)\n"
            "    alphainit = start value of the original objective function fraction (1.0)\n"
            "    alphaquot = redunction factor of alpha (0.9)\n"
            "    alphadist = difference up to which two alpha values are considered equal (0.005)\n"
	    "    epsfeas   = feasibility threshold in order to decide if a new point is (or not) feasible (lb <= Ax <= ub) \n"
	    "    max_time = max time for heuristic FP\n"
	    "    solver = solver used to find analytic center (1=PCx, 2=GLPK, 3=CPLEX)\n"
	    "    norminf = n = standard FP algorithm if no feasible solution between RP-AC, y= infinity norm (integer point less infeasible)\n"
	    "\n"
            "    e.g.: feaspump 10000 2000 70 600 20 .02 s a 4 n n -1 1.0 0.9 0.005 1.0e-6 7200 3 y <lpmodel>\n"
            );
		
      maxIter1 = atoi(argv[1]);
      maxIter2 = atoi(argv[2]);
      int maxit1wi = atoi(argv[3]);
      int maxit2wi = atoi(argv[4]);
      minChange = atoi(argv[5]);
      trsmin = atof(argv[6]);	
      if( *argv[7] == 'r' ) 
         trsrnd = 1;
      else if( *argv[7] == 'h' ) 
         trsrnd = 2;
      else if( *argv[7] == 'q' ) 
         trsrnd=3;
      else if( *argv[7] == 'a' ) 
         trsrnd=4;
      else 
         trsrnd=0;
      bool primal = *argv[8] == 'p';
      bool barrier = *argv[8] == 'b';
      bool autoalg = *argv[8] == 'a';
      int s3me = atoi(argv[9]);	
      bool presolve = *argv[10] == 'y';
      bool mippresolve = *argv[11] == 'y';
      imp = atof(argv[12]);
      double alpha_start = atof(argv[13]);
      double alpha_quot = atof(argv[14]);
      alphadist = atof(argv[15]);
      double epsfeas=atof(argv[16]);
      double max_time=atof(argv[17]);
      int solver = atoi(argv[18]);
      bool norminf = *argv[19] == 'y';
      lpfile = argv[20];

//    aggressive scaling (added Jordi Castro)
//      cplex.setParam(IloCplex::ScaInd, 1);
//    markowitz tolerances (added Jordi Castro)
//      cplex.setParam(IloCplex::EpMrk, 0.3);
//    feasibility tolerance: instrumental to avoid binary variables
//    too far from 0 or 1 in linear relaxations for ill conditioned problems:
//    set to the minimum allowd value 1.0e-9 (added Jordi Castro)
//      cplex.setParam(IloCplex::EpRHS, 1.0e-9);


      modelfile= lpfile;

	
      epInt = cplex.getParam(IloCplex::EpInt);
      cout << "Problem: " << lpfile << endl;
      cout << "Max iter Stage 1: " << maxIter1 << endl;
      cout << "Max iter Stage 2: " << maxIter2 << endl;
      cout << "Min change: " << minChange << endl;
      cout << "Initial Presolve: " << (presolve ? "Yes" : "No") << endl;

      cplex.setParam(IloCplex::PreInd, mippresolve);
                 
      cplex.importModel(model, modelfile, activeObj, vars, rngs);
      cplex.extract(model);
      
      if( mippresolve ) {
	  cout << "applying MIP presolve..." << endl;	  
	  cplex.presolve(IloCplex::NoAlg);
	  CPXENVptr cpxenv = cplex.getImpl()->getCplexEnv();
	  CPXLPptr cpxlp = cplex.getImpl()->getCplexLp();
 	  int restat= CPXgetobjoffset(cpxenv,cpxlp,&obj_offset);
	  if( restat == 0 ) {
	      cout << "presolving objective offset = " << obj_offset << " (restat: " << restat << ")" << endl;
	  }
	  else {
	      cout << "no presolving performed\n" << endl;  
	  }
      }

      auxcplex.importModel(auxmodel, modelfile, auxobj, auxvars, auxrngs);
      auxcplex.extract(auxmodel);

//performing twice presolve is a bit inefficient, but it preserves variable 
//names and variables indices when retrieving solution 

      auxcplex.setOut(env.getNullStream());
      if( mippresolve ) {
	  auxcplex.presolve(IloCplex::NoAlg);
      }

      cplex.setOut(env.getNullStream());
      cplex.setParam(IloCplex::PreInd, presolve);
      auxcplex.setParam(IloCplex::PreInd, presolve);

      nRows = rngs.getSize();
      nVars = vars.getSize();
      cout << "Imported " << nRows << " rows and " << nVars << " columns" << endl;
      cout << "Objective sense: " << (activeObj.getSense() == IloObjective::Maximize ? "Maximize" : "Minimize") << endl;
      cout << "Problem is MIP: " << (cplex.isMIP() ? "Yes" : "No") << endl;
      if( !autoalg ) 
      {
         if( barrier ) 
         { 
            cplex.setParam(IloCplex::RootAlg, IloCplex::Barrier);
            cplex.setParam(IloCplex::BarEpComp,1.0e-8); 			
            cplex.setParam(IloCplex::BarCrossAlg, -1);			
         }
         else if( primal ) 
            cplex.setParam(IloCplex::RootAlg, IloCplex::Primal);
         else 
            cplex.setParam(IloCplex::RootAlg, IloCplex::Dual);
      }
      cout << "Initial Algorithm: " << (autoalg ? "Auto" : (primal ? "Primal" : (barrier ? "Barrier" : "Dual"))) << endl;				


      //cplex 12
	cplex.setParam(IloCplex::Threads,1);

      // save original objective
      origObjExpr = activeObj.getExpr();

      origobj_norm = 0.0;
      IloExpr::LinearIterator	it = origObjExpr.getLinearIterator();
      while( it.ok() )
      {
         origobj_norm += it.getCoef()*it.getCoef();
         ++it;
      }
      origobj_norm = sqrt(origobj_norm);
      alpha = alpha_start;

      IloObjective::Sense objSense = activeObj.getSense();

      // find integer variables
      findIntVars();
      cout << nIntVars << " integer variables ( " << nBinVars << " of which are binary and " << nIntVars-nBinVars << " are general integer)" << endl;
      cout << nVars-nIntVars << " continuous variables" << endl;
      for( int i = 0; i < nIntVars; i++ ) 
      {
         int len = strlen(getVarName(intVars[i]));
         if( maxnamelen < len ) 
            maxnamelen = len;
      }

      // relax model
      IloConversion relaxConv = IloConversion(env, vars, IloNumVar::Float);
      IloConversion auxrelaxConv = IloConversion(env, auxvars, IloNumVar::Float);
      model.add(relaxConv);		
      auxmodel.add(auxrelaxConv);		
      roundedIntVars.resize(nIntVars);
      
     
      prevRoundedIntVars.resize(nIntVars);

      IloRange impcstr(env,-dbl_infinity, origObjExpr);

      deltaVars = new IloNumVar* [nIntVars];
      deltaCnstrUB = new IloRange* [nIntVars];
      deltaCnstrLB = new IloRange* [nIntVars];

      int minDeltaIter = 0;
      vector<double> bestpoint;
      

	info_pos_of_each_variable(nVars);
      
     	if (!(analcent= (double *)calloc(nVars,sizeof(double)))) 
     	{
    		fprintf(stderr, "calloc failed\n"); 
    		exit(-1); 
     	}
     
     	if (!(InitialRP= (double *)calloc(nVars,sizeof(double)))) 
     	{
    		fprintf(stderr, "calloc failed\n"); 
    		exit(-1); 
     	}
	
	if (!(InteriorPoint= (double *)calloc(nVars,sizeof(double)))) 
     	{
    		fprintf(stderr, "calloc failed\n"); 
    		exit(-1); 
     	}
	
	if (!(ACroundedIntVars= (double *)calloc(nIntVars,sizeof(double)))) 
     	{
    		fprintf(stderr, "calloc failed\n"); 
    		exit(-1); 
     	}
	
	
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// finding the analytic center (AC) of the current relaxation, using barrier method without crossover (interior point method) 
// with PCx, GLPK or CPLEX Solver.
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////	
      
     	
	cout << "Finding the analytic center of the current relaxation..." << endl;
	stageAC_time_start = time(NULL);
	

	findAC(analcent,solver);
	stageAC_time_end = time(NULL);
	cout << "TIME TO STAGE AC: " << stageAC_time_end - stageAC_time_start << endl;
	
	isFeasible(analcent,nVars,nRows,epsfeas);
     	
	cout << "VALUE FOBJ WITH AC: " << getFObjValue(analcent) << endl;
	activeObj.setExpr(origObjExpr); // we restart original f.obj
	currentinfeas=-numeric_limits<double>::infinity();
	lbinfeas=numeric_limits<double>::infinity();
	
	cout << "...........done" << endl;

      
      double gamma;
      
      time_t global_time=time(NULL);
      time_t partial_time;

      do {
         ext_iter++; // external iteration count
         s1restarts = 0;
         for( int i = 0; i < nIntVars; i++ ) 
            deltaVars[i] = 0;
         
	 
	 
	 if( ext_iter == 1 ) 
         { 
            
	    // obtain initial relaxed point
            cout << "Solving relaxed problem..." << flush;		
            cplexSolve();
	    lastObj = cplex.getValue(origObjExpr);
	    
	    stage0_time_start = time(NULL);
	    

	    //save the original relaxation in order to can continue if there isn't any feasible solution with gamma [0,1)
	    for (int i=0;i < nVars;i++) {InitialRP[i]=relaxedVars[i];} 
	    
	    
	    for (gamma=0; gamma < (1+epsfeas); gamma=gamma+0.05)
	    {
	    	Point_AC_RP(gamma,analcent);
		roundAC(InteriorPoint,InitialRP,ACroundedIntVars);
	   	if (isFeasible(InteriorPoint,nVars,nRows,epsfeas)) 
		{
			bestgamma=gamma;
			found=true;
			
		}
	    	//check if this infeasibility is better than the previous infeasibility point.
		else
		{
			if (currentinfeas < lbinfeas) {lbinfeas=currentinfeas; lessgamma=gamma;} 	
		}
	    currentinfeas=-numeric_limits<double>::infinity();
	    } // end for
	    
	   if (!found)
	   {
	    	lbinfeas=numeric_limits<double>::infinity();
	    	//use point with better gamma (point more less infeasible)
	    	if (!norminf) Point_AC_RP(1,analcent); //original FP is simulated
		else Point_AC_RP(lessgamma,analcent);
	   }
		
	   lessgamma=0;
	  
	    
STAGE0:	    stage0_time_end = time(NULL);
	    cout << "TIME TO STAGE 0: " << stage0_time_end - stage0_time_start << endl;
	    if (found) 
	    {
	    	Point_AC_RP(bestgamma,analcent);
		roundAC(InteriorPoint,InitialRP,ACroundedIntVars);
		cout << " Stage 0: Finish with gamma (the best): " << bestgamma << " and iter: " << iter << " and value obj= : " << getFObjValue(InteriorPoint) << endl;
		exit(0);
	    }
	    cout << setw(1) << stage << "-" << setw(5) << iter << "           -1 " << setw(3)
                 << restarts << " " << setw(12) << lastObj + obj_offset << setw(12) << getFractionality() << endl;
            cout << "done obj=" << lastObj << " + " << obj_offset << " = " << lastObj + obj_offset << endl;
            lp_bound = lastObj;
            
	    // round bin vars
            round();
            cout << "done" << endl;
            cplex.setParam(IloCplex::RootAlg, IloCplex::Primal);
            bestpoint = roundedIntVars;
            found = isCoincidentS2();
         }
         else { 
            // add improvement constraint
            double imp_obj = best_sol+(lp_bound-best_sol)*imp/100;
            if( objSense == IloObjective::Maximize ) 
               impcstr.setLB(imp_obj);
            else impcstr.setUB(imp_obj);
            if( ext_iter == 2 ) 
            {
               model.add(impcstr);
               rngs.add(impcstr);
            }
            cout << "New Obj cutoff=" << imp_obj << " + " << obj_offset << " = " << imp_obj + obj_offset << endl;
            // reset aux problem var bounds
            for( int i = 0; i < nIntVars; i++ ) 
            {
               auxintvars[i].setUB(intVars[i].getUB());
               auxintvars[i].setLB(intVars[i].getLB());
            }
            found = false;
         }

         stage1_time_start = time(NULL);
         if( !found && nBinVars > 0 && maxIter1 > 0 ) 
         {
            stage = 1;
            cout << "\nStage " << stage << "..." << endl;
            int missedDecr = 0;
            int maxMissedDecr = maxit1wi;
            double minDelta = numeric_limits<double>::infinity();
            // stage 1 : pump only variables at bounds
            while( iter < maxIter1 ) 
            {	
               
               partial_time=time(NULL);
	       if ((partial_time - global_time) > max_time)
	       {
			cout << "MAX TIME OVER: " << partial_time - global_time << endl;
			exit(0);
	       }
               iter++;			
               if( !(iter%20) ) 
                  cout << "s  iter        delta rst          obj    fractionality  incumbent" << endl;

               alpha *= alpha_quot;

               setNewObj();
               cplex_time_start = time(NULL);
               cplexSolve();
	       cplex_time_end = time(NULL);
               cplex_time += cplex_time_end - cplex_time_start;
               
	       double fodelta = foconst + cplex.getValue(dist); 
               prevRoundedIntVars = roundedIntVars;
	       
	    
	    for (int i=0;i < nVars;i++) {InitialRP[i]=relaxedVars[i];} //save the original relaxation
	    
	    for (gamma=0; gamma < (1+epsfeas); gamma=gamma+0.1)
	    {
	    	Point_AC_RP(gamma,analcent);
		roundAC(InteriorPoint,InitialRP,ACroundedIntVars);
	   	if (isFeasible(InteriorPoint,nVars,nRows,epsfeas)) 
		{
			cout << " Stage 1: Finish with gamma: " << gamma << " and iter: " << iter << " and value obj= : " << 			getFObjValue(InteriorPoint) << endl; 
			found=true;
			goto STAGE1;
		}// end if
	   	    	//check if this infeasibility is better than the previous infeasibility point.
		else
		{
			if (currentinfeas < lbinfeas) {lbinfeas=currentinfeas; lessgamma=gamma;} 	
		}
	    currentinfeas=-numeric_limits<double>::infinity();
	    } // end for
	    
	    
	    lbinfeas=numeric_limits<double>::infinity();
	    //use point with better gamma (point more less infeasible)
	    if (!norminf) Point_AC_RP(1,analcent); //original FP is simulated
	    else Point_AC_RP(lessgamma,analcent);

	    lessgamma=0;
   
	       round();	
	       if( fodelta < minDelta ) 
               {
                  cout<<"improved distance by "<<fodelta/minDelta<< "percent"<<endl;
                  if( fodelta/minDelta < 0.9 ) 
                     missedDecr = 0;
                  minDelta = fodelta;				
                  minDeltaIter = iter;
                  bestpoint = roundedIntVars;
               }
               else 
                  missedDecr++;
               lastObj = cplex.getValue(origObjExpr);
               cout << setw(1) << stage << "-" << setw(5) << iter << " " << setw(12) << fodelta << " " << setw(3)
                    << restarts << " " << setw(12) << lastObj + obj_offset << setw(12) << getFractionality();
               if( tot_found ) 
                  cout << " " << setw(12) << best_sol + obj_offset << endl;
               else cout << endl;
               if( fodelta < epInt || isCoincident() ) 
               {
                  cout << endl << "Binary variables didn't change" << endl;
                  break;
               }				
               if( missedDecr > maxMissedDecr && maxIter2 > 0 ) 
               {
                  cout << endl << "Too many iteration without 10% improvement" << endl;
                  break;
               }				
               // during improvement phase, check for feasibility without improvement
               if( ext_iter > 1 ) 
               { 
                  // check feasibility
                  for( int i = 0; i < nIntVars; i++ ) 
                  {
                     if( isBin[i] ) 
                     {
                        auxintvars[i].setUB(roundedIntVars[i]);
                        auxintvars[i].setLB(roundedIntVars[i]);
                     }
                  }
                  int feas = auxcplex.solve();
                  if( feas ) 
                  {
                     double iobj = auxcplex.getObjValue();
                     if( (objSense == IloObjective::Maximize && iobj > lastObj) 
                        || ((objSense == IloObjective::Minimize && iobj < lastObj)) ) 
                     {
                        lastObj = iobj;
                        cout << endl << "Binary variables are feasible" << endl;
                        break;
                     }
                  }
               }
               getNextIntegerPt();
               while( !history.insert(pair<vector<double>,double>(roundedIntVars,alpha)).second && restarts < 10*maxIter1 ) 
               {
                  restart();
               }
            }
            found = isCoincidentS2();
            cout << "Total stage 1 restarts: " << restarts << endl;
            history.clear();
         }			
STAGE1:  stage1_time_end = time(NULL);
         cout << "TIME TO STAGE 1: " << stage1_time_end - stage1_time_start << endl;
	 if (found) exit(0);
	 
	 stage2_time_start = time(NULL);			
         if( !found && maxIter2 > 0 ) 
         {	
            stage = 2;
            cout << "\nStage " << stage << "..." << endl;			

            double minDelta = numeric_limits<double>::infinity();
            // stage 2 : pump all integer variables adding new vars & constraints	
            int s2iterlimit = maxIter2+iter;
            int s2firstiter = iter+1;
            roundedIntVars = bestpoint;
            s1restarts = restarts;
            int missedDecr = 0;
            int maxMissedDecr = maxit2wi;
            cout << "Using best point from iter: " << minDeltaIter << endl;
            while( iter < s2iterlimit ) 
            {
               
		partial_time=time(NULL);
		if ((partial_time - global_time) > max_time)
		{
			cout << "MAX TIME OVER: " << partial_time - global_time << endl;
			exit(0);
		}
		iter++;				
               if( !(iter%20) ) 
                  cout << "s  iter        delta rst          obj    fractionality incumbent" << endl;

               alpha *= alpha_quot;
               setNewObjS2();

               cplex_time_start = time(NULL);
               cplexSolve();
	       
	    for (int i=0;i < nVars;i++) {InitialRP[i]=relaxedVars[i];} //save the original relaxation
	    
	    for (gamma=0; gamma < 1+epsfeas; gamma=gamma+0.1)
	    {
	    	Point_AC_RP(gamma,analcent);
	    	roundAC(InteriorPoint,InitialRP,ACroundedIntVars);
	    	if (isFeasible(InteriorPoint,nVars,nRows,epsfeas)) 
		{
			cout << " Stage 2: Finish with gamma: " << gamma << " and iter: " << iter << " and value obj= : " << 			getFObjValue(InteriorPoint) << endl;
			found=true;
			goto STAGE2;
		} // end if
		//check if this infeasibility is better than the previous infeasibility point.
		else
		{
			if (currentinfeas < lbinfeas) {lbinfeas=currentinfeas; lessgamma=gamma;} 	
		}
	    currentinfeas=-numeric_limits<double>::infinity();
	    } // end for
	    
	    
	    lbinfeas=numeric_limits<double>::infinity();
	    //use point with better gamma (point more less infeasible)
	    
	    if (!norminf) Point_AC_RP(1,analcent); // original FP is simulated 
 	    else Point_AC_RP(lessgamma,analcent);            
	    
	    cplex_time_end = time(NULL);
            cplex_time += cplex_time_end - cplex_time_start;				

            double fodelta = foconst + cplex.getObjValue(); 
            prevRoundedIntVars = roundedIntVars;
            round();
            if( fodelta < minDelta ) 
            {
                  if( fodelta/minDelta < 0.9 ) 
                     missedDecr = 0;
                  minDelta = fodelta;				
                  minDeltaIter = iter;
                  bestpoint = roundedIntVars;
            }
            else 
                  missedDecr++;
               lastObj = cplex.getValue(origObjExpr);
               cout << setw(1) << stage << "-" << setw(5) << iter << " " << setw(12) << fodelta << " " << setw(3)
                    << restarts << " " << setw(12) << lastObj + obj_offset << setw(12) << getFractionality();
               if( tot_found ) 
                  cout << " " << setw(12) << best_sol + obj_offset << endl;
               else 
                  cout << endl;
               if( isCoincidentS2() ) 
               {
                  found = true;
                  break;
               }
               if( missedDecr > maxMissedDecr ) 
               {
                  cout << endl << "Too many iteration without 10% improvement" << endl;
                  break;
               }
               if( restarts-s1restarts >= 100 ) 
               {
                  cout << endl << "Too many restarts" << endl;
                  break;
               }
               // during improvement phase, check for feasibility without improvement
               if( ext_iter > 1 ) 
               { 
                  // check feasibility
                  for( int i = 0; i < nIntVars; i++ ) 
                  {
                     auxintvars[i].setUB(roundedIntVars[i]);
                     auxintvars[i].setLB(roundedIntVars[i]);
                  }
                  int feas = auxcplex.solve();
                  if( feas ) 
                  {
                     double iobj = auxcplex.getObjValue();
                     if( (objSense == IloObjective::Maximize && iobj > lastObj) 
                        || ((objSense == IloObjective::Minimize && iobj < lastObj)) ) 
                     {
                        lastObj = iobj;
                        cout << endl << "Integer variables are feasible" << endl;
                        found = true;
                        break;
                     }
                  }
               }
               getNextIntegerPtS2();
               while( !history.insert(pair<vector<double>,double>(roundedIntVars,alpha)).second 
                  && restarts-s1restarts < 100) 
                  restartS2();
               while( history.size() > 2000 ) 
                  history.erase(history.begin());
            }			
            cout << "Total stage 2 restarts: " << (restarts - s1restarts) << endl;
            history.clear();
         }
STAGE2:  stage2_time_end = time(NULL);
	 cout << "TIME TO STAGE 2: " << stage2_time_end - stage2_time_start << endl;
	 if (found) exit(0);
	 
	 stage3_time_start = time(NULL);
         if( !found ) 
         { 
            // enumerate
            stage = 3;
            cout << "\nStage " << stage << "..." << endl;
            roundedIntVars = bestpoint;
            cout << "Using best point from iter: " << minDeltaIter << endl;
            // make problem MIP again
            model.remove(relaxConv);
            cplex.setParam(IloCplex::IntSolLim, 1);
            cplex.setParam(IloCplex::MIPEmphasis, s3me);
            cplex.setParam(IloCplex::EpRHS, cplex.getDefault(IloCplex::EpRHS));
            cplex.setOut(env.out());
            setNewObjS2();	
            cout << "obj offset: " << foconst << " + " << obj_offset << " = " << foconst + obj_offset << endl;
            cplexSolve();
            double newlp_bound = cplex.getBestObjValue();
            cout << "Status: " << cplex.getStatus() << endl;
            cout << "CplexStatus:" << cplex.getCplexStatus() << endl;
            found = cplex.isPrimalFeasible() == 1;
         }
         stage3_time_end = time(NULL);
	 
	 lastObj = cplex.getValue(origObjExpr);
         if( ext_iter == 1 ) 
         {
            first_iter = iter;
            first_sol = lastObj;
            first_restarts = restarts;
            first_stage = stage;
         }
         if( found ) 
         {					
            tot_found++;
            // if found in stage 3, retrieve int vars
            if( stage == 3 )  
            {
               cplex.getValues(relaxedIntVars, intVars);
               for( int i = 0; i < nIntVars; i++ ) 
                  roundedIntVars[i] = floor(relaxedIntVars[i]+.5);
               if( imp >= 0 ) 
               {
                  cplex.setOut(env.getNullStream());
                  model.add(relaxConv);
               }
            }
            cout << "Solution with obj=" << lastObj << " + " << obj_offset << " = " << lastObj + obj_offset << " found" << endl;

            // save MST file
            string mstname(lpfile);
            string::size_type dotpos = mstname.find_last_of('.');
            if( dotpos != string::npos ) 
               mstname.resize(dotpos);
            mstname += ".mst";
            cout << "Writing MIP start values to file " << mstname << endl;
            ofstream mst(mstname.c_str());
            mst << "NAME" << endl;
            for( int i = 0; i < nIntVars; i++ ) 
               mst << "     " << setiosflags(ios_base::left) << setw(maxnamelen) 
                   << getVarName(intVars[i]) << " " << roundedIntVars[i]  << "\n";
            mst << "ENDATA\n";
            mst.close();
				
            if( imp >= 0 )  
            { 
               // try fixing general variable and reoptimize
               cout << "Trying to improve..." << flush;
               for( int i = 0; i < nIntVars; i++ ) 
               {
                  auxintvars[i].setUB(roundedIntVars[i]);
                  auxintvars[i].setLB(roundedIntVars[i]);
               }
               bool fea = auxcplex.solve() == 0;
               double iobj = auxcplex.getObjValue();
               if( fea && ((objSense == IloObjective::Maximize && iobj > best_sol) 
                     || ((objSense == IloObjective::Minimize && iobj < best_sol)))) 
               {
                  cout << "found." << endl;
                  lastObj = iobj;
                  cout << "Solution with obj= " << lastObj << " + " << obj_offset << " = " << lastObj + obj_offset << " found" << endl;
               }
               else 
                  cout << "not found." << endl;
            }
            if( ext_iter == 1 || (objSense == IloObjective::Maximize && best_sol < lastObj) 
               || (objSense == IloObjective::Minimize && best_sol > lastObj) ) 
               best_sol = lastObj;
         }
         tot_iter += iter;
         tot_restarts += restarts;
         iter = 0;
         restarts = 0;
         if( fabs((lp_bound-best_sol)/lp_bound) < 0.001 ) 
         {
            cout << "Gap within 0.1%, exiting" << endl;
            break;
         }
      } while( imp > 0 && found );
      
	
      cout << "TIME TO STAGE 3: " << stage3_time_end - stage3_time_start << endl;

      timeend = time(NULL);
   }
   catch(IloException& e) {
      cout << "Concert exception: " << e << endl; 
      try 
      {
         cout << "Last cplex status: " << cplex.getStatus() << "-" << cplex.getCplexStatus() << endl;
      } 
      catch(...) {};
   }
   catch(string& e) {cerr << e << endl; }
   catch(...) {cerr << "Unknown exception caught" << endl; }
   // report result		
   if( lpfile ) 
   {
      // display solution
      cout << "Solution (only non-zero entries are reported):" << endl;
      cout << "obj                            = " << setw(16) << first_sol + obj_offset << endl;
      for( int i = 0; i < nIntVars; i++ )
      {
         if( fabs(roundedIntVars[i]) > 1e-09 )
            cout << setw(30) << setiosflags(ios_base::left) << getVarName(intVars[i]) << resetiosflags(ios_base::left)
                 << " = " << setw(16) << roundedIntVars[i] << endl;
      }

      tot_iter += iter;
      tot_restarts += restarts;
      setprecision(12);
    
      cout << "Feasible " << (tot_found ? "FOUND" : "NOT found") << " in " << (tot_found ? first_iter : tot_iter) << " iterations!" << endl;
      if( tot_found ) 
         cout << "First sol: obj= " << setprecision(16) << first_sol + obj_offset << " time= " << timeend - timestart 
              << " iter= " << first_iter << " restarts= " << first_restarts << " stage= " << first_stage << endl;
      if( imp > 0 ) 
         cout << "Best sol: obj= " << setprecision(16) << best_sol + obj_offset << " found= " << tot_found << " iter= " 
              << tot_iter << " restarts= " << tot_restarts << " stage= " << stage << endl;
   }
   env.end();

   // remove temporary presolving file
   if( removetmppresolfile )
      unlink(tmppresolfile);
} 

